open Stack

type status = {
  stt : bool;
  sb : bool;
  si : bool;
  sem : bool;
  ss : bool;
  su : int;
  sc : char;
  ssz : char;
}

let test b = if b then "1" else "0"

let print_char = function
  | '\013' -> print_string "^M"
  | '\010' -> print_string "^A"
  | '\009' -> print_string "^I"
  | c -> print_char c; print_char ' '
    
let print_status s = (* add color, size *)
  Printf.printf " B=%s EM=%s TT=%s I=%s S=%s U=%d S=%c C=%c\n"
    (test s.sb)
    (test s.sem)
    (test s.stt)
    (test s.si)
    (test s.ss)
    s.su
    s.ssz
    s.sc

let initial_color = 'X'
let initial_size = 'X'
let initial_status = {
  stt = false;
  sb = false;
  si = false;
  sem = false;
  ss = false;
  su = 0;
  sc = initial_color;
  ssz = initial_size
}

let is_strong_set s = s.ss
let is_tt_set s = s.stt
let get_under s = s.su

let set_tt s = {s with stt = true }
let set_bold s = {s with sb = true }
let set_strong s = {s with ss = true }
let set_italic s = {s with si = true }

let unset_tt s = {s with stt = false }
let unset_bold s = {s with sb = false }
let unset_strong s = {s with ss = false }
let unset_italic s = {s with si = false }
let unset_em s = {s with sem = false }
let reset_under s = {s with su = 0 }

let toggle_em s = {s with sem = not s.sem }
let incr_underline s = {s with su = s.su + 1 }
let decr_underline s = {s with su = s.su - 1 }

let set_color s c = {s with sc = c }
let set_size s sz = {s with ssz = sz }

let space_filter s =
  { s
  with
    sb = false; si = false; sem = false; ss = false;
    sc = if s.su = 0 then 'w'else s.sc }

let status_stack = ref []
let push_status c = status_stack := c :: !status_stack
let top_status () = List.hd !status_stack
let pop_status () =
  match !status_stack with s::st -> status_stack := st; s | _ -> failwith "s"

let plain_open_action status =
  push_status status;
  let status = reset_under status in
  let status = unset_strong status in
  let status = unset_em status in
  let status = unset_italic status in
  let status = unset_bold status in
  unset_tt status

let plain_close_action status = pop_status ()

let bold_open_action status =
  push_status status;
  set_bold status

let bold_close_action status = pop_status ()

let tt_open_action status =
  push_status status;
  set_tt status

let tt_close_action status = pop_status ()

let emph_open_action status =
  push_status status;
  if not (is_strong_set status) then toggle_em status else status

let emph_close_action status = pop_status ()

let strong_open_action status =
  push_status status;
  unset_em (set_strong status)

let strong_close_action status = pop_status ()

let italic_open_action status =
  push_status status;
  set_italic status

let italic_close_action status = pop_status ()

let underline_open_action status =
  push_status status;
  if get_under status < 3 then incr_underline status else status

let underline_close_action status = pop_status ()

let color_open_action c status =
  push_status status;
  set_color status c

let color_close_action c status = pop_status ()

let size_open_action s status =
  push_status status;
  set_size status s

let size_close_action s status = pop_status ()

let look_ahead () = input_char stdin

let parse_tag status =
  let tag =
    match look_ahead () with
      | 'B' -> bold_open_action status 
      | 'E' ->
	  (match look_ahead () with
	    | 'M' -> emph_open_action status 
	    | _ -> failwith "Syntax error: M expected")
      | 'I' -> italic_open_action status 
      | 'P' ->
	  (match look_ahead () with
	    | 'L' -> plain_open_action status 
	    | _ -> failwith "Syntax error: L expected")
      | 'S' -> strong_open_action status 
      | 'T' ->
	  (match look_ahead () with
	    | 'T' -> tt_open_action status 
	    | _ -> failwith "Syntax error: T expected")
      | 'U' -> underline_open_action status 
      | ('0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9') as s ->
	  size_open_action s status 
      | ('r'|'g'|'b'|'c'|'m'|'y'|'k'|'w') as c ->
	  color_open_action c status 
      | '/' -> begin
	  match look_ahead () with
	    | 'B' -> bold_close_action status 
	    | 'E' ->
		(match look_ahead () with
		  | 'M' -> emph_close_action status 
		  | _ -> failwith "Syntax error: M expected")
	    | 'I' -> italic_close_action status 
	    | 'P' ->
		(match look_ahead () with
		  | 'L' -> plain_close_action status 
		  | _ -> failwith "Syntax error: L expected")
	    | 'S' -> strong_close_action status 
	    | 'T' ->
		(match look_ahead () with
		  | 'T' -> tt_close_action status 
		  | _ -> failwith "Syntax error: T expected")
	    | 'U' -> underline_close_action status 
	    | ('0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9') as s ->
		size_close_action s status 
	    | ('r'|'g'|'b'|'c'|'m'|'y'|'k'|'w') as c ->
		color_close_action c status 
	    | _ -> failwith "Syntax error: tag expected"
	end
      | _ -> failwith "Syntax error: tag expected"
  in
  match look_ahead () with
    | '>' -> tag
    | _ -> failwith "Syntax error: '>' expected"

(* Main parsing function *)
let rec parse_action last_space_status status =
  match look_ahead () with
    | '<' -> (* Should come first *)
	parse_action last_space_status (parse_tag status)
    | ('\033'..'\127') as c ->
	print_char c; print_status status;
	parse_action None status
    | (' ' | '\013' | '\010' | '\009') as c ->
	let space_status = space_filter status in
	if is_tt_set space_status then
	  begin
	    print_char c; print_status space_status;
	    parse_action (Some space_status) status
	  end
	else
	  (match last_space_status with
	    | None ->
		print_char ' '; print_status space_status;
		parse_action (Some space_status) status
	    | Some s ->
		parse_action last_space_status status)
    | _ -> failwith "Syntax error: unexpected char"

let semantics () = parse_action None initial_status

let _ = if not (!Sys.interactive) then
  try semantics ()
  with End_of_file -> ()
