open Transition
type color = int

let color_of_token = function
  | NOTAG -> 0
  | CRO -> 1 lsl 4
  | CGO -> 2 lsl 4
  | CBO -> 3 lsl 4
  | CCO -> 4 lsl 4
  | CMO -> 5 lsl 4
  | CYO -> 6 lsl 4
  | CKO -> 7 lsl 4
  | CWO -> 8 lsl 4
  | UNDEFINED -> 9 lsl 4
  | _ -> failwith "Not a color"

let white_color = color_of_token CWO

let char_of_color = function
  |  16 -> 'r'
  |  32 -> 'g'
  |  48 -> 'b'
  |  64 -> 'c'
  |  80 -> 'm'
  |  96 -> 'y'
  | 112 -> 'k'
  | 128 -> 'w'
  | 144 -> 'X'
  | c -> (* failwith "Not a defined color"*) print_int c; print_string" "; '?'

type size = int

let size_of_token = function
  | NOTAG -> 0
  | S0O -> 1 lsl 8
  | S1O -> 2 lsl 8
  | S2O -> 3 lsl 8
  | S3O -> 4 lsl 8
  | S4O -> 5 lsl 8
  | S5O -> 6 lsl 8
  | S6O -> 7 lsl 8
  | S7O -> 8 lsl 8
  | S8O -> 9 lsl 8
  | S9O -> 10 lsl 8
  | UNDEFINED -> 11 lsl 8
  | _ -> failwith "Not a size"

let char_of_size = function
  |  256 -> '0'
  |  512 -> '1'
  |  768 -> '2'
  | 1024 -> '3'
  | 1280 -> '4'
  | 1536 -> '5'
  | 1792 -> '6'
  | 2048 -> '7'
  | 2304-> '8'
  | 2560 -> '9'
  | 2816 -> 'X'
  | c -> (* failwith "Not a defined size"*)  print_int c; print_string" ";'?'

type status = int

let copy_status s = s

(* Format for status *)
(*  ___  ___r  __tt bbii  ssee zzzz  cccc uuuu *)
(* r = remove tag *)

let is_tt_set s     = (s land 0x00300000) <> 0
let is_bold_set s   = (s land 0x000C0000) <> 0
let is_italic_set s = (s land 0x00030000) <> 0
let is_strong_set s = (s land 0x0000C000) <> 0

let is_em_set s     = (s land 0x00003000) <> 0
let is_plain_set s  = (s land 0x003FF00F) = 0

let get_underline s = (s land 0x0000000F)
let get_color s     = (s land 0x000000F0)
let get_size s      = (s land 0x00000F00)

let set_tt s        = s lor 0x00300000
let set_bold s      = s lor 0x000C0000
let set_italic s    = s lor 0x00030000
let set_strong s    = s lor 0x0000C000

let unset_tt s      = s land 0xFFCFFFFF
let unset_bold s    = s land 0xFFF3FFFF
let unset_italic s  = s land 0xFFFCFFFF
let unset_strong s  = s land 0xFFFF3FFF

let toggle_em s      = s lxor 0x00003000
let incr_underline s = s + 1 (* unsafe !! *)
let decr_underline s = s - 1 (* unsafe !! *)
let set_plain      s = s land 0xFFC00FF0

let set_color s c    = (s land 0xFFFFFF0F) lor c
let set_size s sz    = (s land 0xFFFFF0FF) lor sz


type base_status = int
let get_base s = s land 0x003FF00F
let set_base s b = s lor b

let test b = if b then "1" else "0"

let print_status s = (* add color, size *)
  Printf.printf " B=%s EM=%s TT=%s I=%s S=%s U=%d S=%c C=%c\n"
    (test (is_bold_set s))
    (test (is_em_set s))
    (test (is_tt_set s))
    (test (is_italic_set s))
    (test (is_strong_set s))
    (get_underline s)
    (char_of_size (get_size s))
    (char_of_color (get_color s))

type space_status = status

let space_filter s = 
  if get_underline s = 0 then
(s land 0xFFF00F0F) lor white_color
 else
 (s land 0xFFF00FFF)

let initial_base = 0
let initial_color = color_of_token UNDEFINED
let initial_size = size_of_token UNDEFINED
let initial_status () =
  set_size (set_color initial_base initial_color) initial_size

let set_remove s    = s lor 0x01000000
let is_remove_set s = (s land 0x01000000) <> 0

(* Format for actions *)
(*  ___  __pp  __tt bbii  ssee zzzz  cccc uuuu *)

module Act = struct
  type action = {
    mutable aplain : bool;
    mutable abase : tag list;
    mutable asize : size;
    mutable acolor : color 
  }

  let empty_action = 
     { aplain = false; abase = []; asize = size_of_token NOTAG; acolor = color_of_token NOTAG }

  let rec add_em_action = function
    |[] -> [EMO]
    |EMO::q -> q
    |t::q -> t::(add_em_action q)

  let is_plain_set act = act.aplain
  let set_plain act = { act with aplain = true }
  let toggle_em act = { act with abase = add_em_action act.abase }
  let set_base act l = { act with abase = l }
  let add_base tk act = { act with abase = tk :: act.abase }
  let unset_plain act = { act with aplain = false }

  let set_size act sz = { act with asize = sz }
  let set_color act c = { act with acolor = c }
  let unset_size act = { act with asize = size_of_token NOTAG }
  let unset_color act = { act with acolor = color_of_token NOTAG }
  let get_size act = act.asize
  let get_color act = act.acolor
  let is_size_defined act = act.asize <> size_of_token NOTAG
  let is_color_defined act = act.acolor <> color_of_token NOTAG


let print_close_token = function
  | BO -> print_string "</B>"
  | EMO -> print_string "</EM>"
  | IO -> print_string "</I>"
  | PLO -> print_string "</PL>"
  | SO -> print_string "</S>"
  | TTO -> print_string "</TT>"
  | UO -> print_string "</U>"
(* Sizes *)
  | S0O -> print_string "</0>"
  | S1O -> print_string "</1>"
  | S2O -> print_string "</2>"
  | S3O -> print_string "</3>"
  | S4O -> print_string "</4>"
  | S5O -> print_string "</5>"
  | S6O -> print_string "</6>"
  | S7O -> print_string "</7>"
  | S8O -> print_string "</8>"
  | S9O -> print_string "</9>"
(* Colours *)
  | CRO -> print_string "</r>"
  | CGO -> print_string "</g>"
  | CBO -> print_string "</b>"
  | CCO -> print_string "</c>"
  | CMO -> print_string "</n>"
  | CYO -> print_string "</y>"
  | CKO -> print_string "</k>"
  | CWO -> print_string "</w>"
(* Special code for removing *)
  | REMOVE | UNDEFINED -> failwith "Warning: undefined"
  | _ -> failwith "Printing unauthorized token"

let print_open_token = function
  | BO -> print_string "<B>"
  | EMO -> print_string "<EM>"
  | IO -> print_string "<I>"
  | PLO -> print_string "<PL>"
  | SO -> print_string "<S>"
  | TTO -> print_string "<TT>"
  | UO -> print_string "<U>"
(* Sizes *)
  | S0O -> print_string "<0>"
  | S1O -> print_string "<1>"
  | S2O -> print_string "<2>"
  | S3O -> print_string "<3>"
  | S4O -> print_string "<4>"
  | S5O -> print_string "<5>"
  | S6O -> print_string "<6>"
  | S7O -> print_string "<7>"
  | S8O -> print_string "<8>"
  | S9O -> print_string "<9>"
(* Colours *)
  | CRO -> print_string "<r>"
  | CGO -> print_string "<g>"
  | CBO -> print_string "<b>"
  | CCO -> print_string "<c>"
  | CMO -> print_string "<n>"
  | CYO -> print_string "<y>"
  | CKO -> print_string "<k>"
  | CWO -> print_string "<w>"
(* Special code for removing *)
  | REMOVE | UNDEFINED -> print_string "Warning: undefined"
  | _ -> failwith "Printing unauthorized token"

let print_open_color c =
  print_char '<';
  print_char (char_of_color c);
  print_char '>'

let print_close_color c =
  print_string "</";
  print_char (char_of_color c);
  print_char '>'

let print_open_size c =
  print_char '<';
  print_char (char_of_size c);
  print_char '>'

let print_close_size c =
  print_string "</";
  print_char (char_of_size c);
  print_char '>'

let print_close_action { aplain=pl; abase=l; asize=sz; acolor=c } =
  List.iter print_close_token l;
  if pl then print_close_token PLO;
  if sz <> size_of_token NOTAG then print_close_size sz;
  if c <> color_of_token NOTAG then print_close_color c

let print_open_action { aplain=pl; abase=l; asize=sz; acolor=c } =
  if c <> color_of_token NOTAG then print_open_color c;
  if sz <> size_of_token NOTAG then print_open_size sz;
  if pl then print_open_token PLO;
  List.iter print_open_token (List.rev l)

end
